#include "header.h"

extern "C" { int _fltused=0; } // no crt so ... float support is questionable...
GLuint g_Texture;
HGLRC g_hRC = NULL;
HDC g_hDC;


HWND g_hWnd = NULL;
DWORD g_dwHeight = 640;
DWORD g_dwWidth = 480;
DWORD g_dwColorDepth = 16;
DWORD g_dwRefreshRate = 60;
IUnknown* g_ddsPrimary = NULL;


void FillPixelFormat( DDPIXELFORMAT* ddpf, DWORD bpp )
{
	ddpf->dwSize = sizeof(DDPIXELFORMAT);
	ddpf->dwRGBBitCount = bpp;

	switch( bpp )
	{
	case 32:
		ddpf->dwBBitMask = 0x000000FF;
		ddpf->dwGBitMask = 0x0000FF00;
		ddpf->dwRBitMask = 0x00FF0000;
		ddpf->dwRGBAlphaBitMask = 0xFF000000;
		ddpf->dwFlags = DDPF_RGB;
		break;
	case 24:
		ddpf->dwBBitMask = 0x000000FF;
		ddpf->dwGBitMask = 0x0000FF00;
		ddpf->dwRBitMask = 0x00FF0000;
		ddpf->dwRGBAlphaBitMask = 0;
		ddpf->dwFlags = DDPF_RGB;
		break;
	case 16: // 565
		ddpf->dwBBitMask = 0x0000001F;
		ddpf->dwGBitMask = 0x000007E0;
		ddpf->dwRBitMask = 0x0000F800;
		ddpf->dwRGBAlphaBitMask = 0;
		ddpf->dwFlags = DDPF_RGB;
		break;
	/* 
	case 16: // 555
		ddpf->dwBBitMask = 0x0000001F;
		ddpf->dwGBitMask = 0x000003E0;
		ddpf->dwRBitMask = 0x00007C00;
		ddpf->dwRGBAlphaBitMask = 0;

	case 8:
		ddpf->dwBBitMask = 0;
		ddpf->dwGBitMask = 0;
		ddpf->dwRBitMask = 0;
		ddpf->dwRGBAlphaBitMask = 0;
		ddpf->dwFlags = DDPF_RGB | DDPF_PALETTEINDEXED8;
		break;
	case 4:
		// DDPF_PALETTEINDEXED4
	*/
	}
}

void SetupOpenGL()
{
	static BOOL firstrun = TRUE;

	if( firstrun == TRUE )
	{
		firstrun = FALSE;
		g_hDC = GetDC( g_hWnd );
		PIXELFORMATDESCRIPTOR pfd =
		{
			sizeof(PIXELFORMATDESCRIPTOR),
			1,
			PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,    //Flags
			PFD_TYPE_RGBA,            //The kind of framebuffer. RGBA or palette.
			32,                        //Colordepth of the framebuffer.
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			0,                        //Number of bits for the depthbuffer
			0,                        //Number of bits for the stencilbuffer
			0,                        //Number of Aux buffers in the framebuffer.
			PFD_MAIN_PLANE,
			0, 0, 0, 0
		};
		SetPixelFormat( g_hDC, ChoosePixelFormat( g_hDC, &pfd ), &pfd );
		g_hRC = wglCreateContext( g_hDC );
		wglMakeCurrent( g_hDC, g_hRC );

		// enable vsync
		// WGL_EXT_swap_control
		PFNWGLSWAPINTERVALEXTPROC wglSwapIntervalEXT = (PFNWGLSWAPINTERVALEXTPROC) wglGetProcAddress( "wglSwapIntervalEXT" );
		if( (long)wglSwapIntervalEXT > 4 )
		{
			wglSwapIntervalEXT( 1 );
		}
	}
	else // not first run
	{
		glDeleteTextures( 1, &g_Texture );
	}

	glClearColor( 0, 0, 0, 0 ); // black
	glClear( GL_COLOR_BUFFER_BIT ); 
	SwapBuffers( g_hDC );
	glViewport( 0, 0, (GLsizei)g_dwWidth, (GLsizei)g_dwHeight );
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho( 0, (double)g_dwWidth, (double)g_dwHeight, 0, -1, 1 );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glEnable( GL_TEXTURE_2D );
	glEnable( GL_DITHER );
	glDisable( GL_LIGHTING );
	glDisable( GL_BLEND );
	glDisable( GL_DEPTH_TEST );

	glGenTextures( 1, &g_Texture );
	glBindTexture( GL_TEXTURE_2D, g_Texture );
	
	// NPOT
	switch( g_dwColorDepth )
	{
	case 16:
		// glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB5_A1, g_dwWidth, g_dwHeight, 0, GL_BGRA, GL_UNSIGNED_SHORT_1_5_5_5_REV, 0 ); // 555
		glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, g_dwWidth, g_dwHeight, 0, GL_RGB,  GL_UNSIGNED_SHORT_5_6_5, 0 ); // 565
		break;
	case 24:
		glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB8, g_dwWidth, g_dwHeight, 0, GL_BGR,  GL_UNSIGNED_BYTE, 0 ); // 24-bit
		break;
	case 32:
		glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, g_dwWidth, g_dwHeight, 0, GL_BGRA, GL_UNSIGNED_BYTE, 0 ); // 32-bit
		break;
	}
	
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
	glPixelStorei( GL_UNPACK_ALIGNMENT, 4 );

}

HRESULT CreateSurfaceHelper( WRAP* This, LPDDSURFACEDESC lpDDSurfaceDesc, LPDIRECTDRAWSURFACE* lplpDDSurface )
{
	HRESULT hResult;

	// if creating primary surface
	if( ( ( lpDDSurfaceDesc->dwFlags & DDSD_CAPS ) || ( lpDDSurfaceDesc->dwFlags == DDSD_ALL ) ) && 
		( lpDDSurfaceDesc->ddsCaps.dwCaps & DDSCAPS_PRIMARYSURFACE ) )
	{
		// v4 or higher required to create surfaces in client allocated memory
		LPDIRECTDRAW4 lpDD;
		if( ( This->xVtbl == &dd::xVtbl7 ) || ( This->xVtbl == &dd::xVtbl4 ) ) {
			This->unknwn->lpVtbl->AddRef( This->unknwn );
			lpDD = This->dd4;
		}
		else {
			if( FAILED( This->dd4->lpVtbl->QueryInterface( This->dd4, IID_IDirectDraw4, (void**)&lpDD ) ) ) {
				lpDD = NULL;
			}
		}
	
		hResult = E_FAIL;
		
		if( lpDD != NULL ) { 
			BYTE* buf; 
			LPDIRECTDRAWSURFACE4 front;
			DDSURFACEDESC2 ddsd;
			DWORD surf_data_size;
			DWORD dwBackBuffers;
			
			// create a fake primary surface
			RtlSecureZeroMemory( &ddsd, sizeof( ddsd ) );
			ddsd.dwSize = sizeof( ddsd );
			ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_PIXELFORMAT | DDSD_LPSURFACE | DDSD_PITCH;
			ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
			ddsd.dwHeight = g_dwHeight;
			ddsd.dwWidth = g_dwWidth;
			ddsd.lPitch = (g_dwWidth * ( g_dwColorDepth >> 3 ) + 3) & ~3;
			FillPixelFormat( &ddsd.ddpfPixelFormat, g_dwColorDepth );
			surf_data_size = ddsd.dwHeight * ddsd.lPitch;
			dwBackBuffers = ( lpDDSurfaceDesc->dwFlags & DDSD_BACKBUFFERCOUNT ) ? lpDDSurfaceDesc->dwBackBufferCount : 0;
			buf = (BYTE*) VirtualAlloc( NULL, surf_data_size * ( 1 + dwBackBuffers ), MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );
			if( buf != NULL ){
				ddsd.lpSurface = buf;
				hResult = lpDD->lpVtbl->CreateSurface( lpDD, &ddsd, (LPDIRECTDRAWSURFACE4*) &front, 0 );
				if( SUCCEEDED( hResult ) ){
					// emulate implicite complex surface creation for the primary surface
					LPDIRECTDRAWSURFACE4 next = NULL;
					LPDIRECTDRAWSURFACE4 prev = front;
					for( DWORD i = 0; i < dwBackBuffers; i++ )
					{
						buf += surf_data_size;
						ddsd.lpSurface = buf;
						lpDD->lpVtbl->CreateSurface( lpDD, &ddsd, &next, 0 );
						prev->lpVtbl->AddAttachedSurface( prev, next );
						next->lpVtbl->Release( next ); 
						prev = next;
					}

					// query for the correct surface interface version
					if( ( This->xVtbl == &dd::xVtbl1 ) || ( This->xVtbl == &dd::xVtbl2 ) ){
						if( SUCCEEDED( front->lpVtbl->QueryInterface( front, IID_IDirectDrawSurface, (void**)lplpDDSurface ) ) ){
							front->lpVtbl->Release( front );
						}
					}
					else{ // already have the correct surface interface version
						*lplpDDSurface = (LPDIRECTDRAWSURFACE) front;
					}

					// store primary surface base object
					(*lplpDDSurface)->lpVtbl->QueryInterface( *lplpDDSurface, IID_IUnknown, (void**)&g_ddsPrimary );
					g_ddsPrimary->lpVtbl->Release( g_ddsPrimary );

					//
					SetupOpenGL();

					hResult = DD_OK;
				}
			}
			lpDD->lpVtbl->Release( lpDD );
		}
	}

	else // not primary surface
	{
		///////// required apparently.... todo...
		lpDDSurfaceDesc->ddsCaps.dwCaps |= DDSCAPS_OFFSCREENPLAIN;
		////////////
		
		if( !( lpDDSurfaceDesc->dwFlags & DDSD_PIXELFORMAT ) )
		{
			lpDDSurfaceDesc->dwFlags |= DDSD_PIXELFORMAT;
			FillPixelFormat( &lpDDSurfaceDesc->ddpfPixelFormat, g_dwColorDepth );
		}
		hResult =  This->dd1->lpVtbl->CreateSurface( This->dd1, lpDDSurfaceDesc, lplpDDSurface, 0 );
		if( hResult == DDERR_NODIRECTDRAWHW ) hResult = DDERR_OUTOFVIDEOMEMORY;
	}
	return hResult;
}

BOOL IsPrimary( IDirectDrawSurface* dds )
{
	BOOL bResult;
	IUnknown* iface;
	dds->lpVtbl->QueryInterface( dds, IID_IUnknown, (void**)&iface );
	bResult = ( iface == g_ddsPrimary ) ? TRUE : FALSE;
	iface->lpVtbl->Release( iface );
	return bResult;
}

void GoFullScreen()
{
	SetWindowPos( g_hWnd, HWND_TOP, 0, 0, g_dwWidth, g_dwHeight, SWP_SHOWWINDOW ); 

	DEVMODE devMode;
	RtlSecureZeroMemory(&devMode, sizeof(devMode));
	devMode.dmSize = sizeof(devMode);
	devMode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT; 
	devMode.dmBitsPerPel = 32;
    devMode.dmPelsWidth  = g_dwWidth;
    devMode.dmPelsHeight = g_dwHeight;
	ChangeDisplaySettings( &devMode, CDS_FULLSCREEN );
}

void ToScreen( WRAP* This )
{
	DDSURFACEDESC2 ddsd;
	if( This->xVtbl == &dd::xVtbl1 || This->xVtbl == &dd::xVtbl2 ) ddsd.dwSize = sizeof( DDSURFACEDESC );
	else ddsd.dwSize = sizeof( DDSURFACEDESC2 );

	RECT rc;
	GetClientRect( g_hWnd, &rc );
	ClientToScreen( g_hWnd, (POINT*)&rc.left );
	ClientToScreen( g_hWnd, (POINT*)&rc.right );

	// clip window
	if( ( rc.top >= (long)g_dwHeight ) ||
	( rc.bottom <= 0 ) ||
	( rc.right <= 0 ) ||
	( rc.left >= (long)g_dwWidth ) )
	{
		return;
	}
	rc.left = max( rc.left, 0 );
	rc.right = min( rc.right, (long)g_dwWidth );
	rc.top = max( rc.top, 0 );
	rc.bottom = min( rc.bottom,(long)g_dwHeight );

	DWORD w = rc.right - rc.left;
	DWORD h = rc.bottom - rc.top;

	if(SUCCEEDED( This->dds1->lpVtbl->Lock( This->dds1, NULL, (DDSURFACEDESC*) &ddsd, DDLOCK_WAIT, NULL) ) )
	{
	
		glPixelStorei( GL_UNPACK_ROW_LENGTH, ddsd.lPitch / ( ddsd.ddpfPixelFormat.dwRGBBitCount / 8 ) );
		glPixelStorei( GL_UNPACK_SKIP_PIXELS, rc.left );
		glPixelStorei( GL_UNPACK_SKIP_ROWS, rc.top );

		switch( g_dwColorDepth )
		{	
			case 16:
				glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, w, h, GL_RGB, GL_UNSIGNED_SHORT_5_6_5, ddsd.lpSurface ); // 565
				// glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, w, h, GL_BGRA, GL_UNSIGNED_SHORT_1_5_5_5_REV, ddsd.lpSurface  ); // 555
				break;
			case 24:
				glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, w, h, GL_BGR,  GL_UNSIGNED_BYTE, ddsd.lpSurface );
				break; 
			case 32:
				glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, w, h, GL_BGRA, GL_UNSIGNED_BYTE, ddsd.lpSurface );
				break;
			*/
		}

		glPixelStorei( GL_UNPACK_ROW_LENGTH, 0 );
		glPixelStorei( GL_UNPACK_SKIP_PIXELS, 0 );
		glPixelStorei( GL_UNPACK_SKIP_ROWS, 0 );


		glViewport( 0, 0, (GLsizei)w, (GLsizei)h );
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho( 0, (double)w, (double)h, 0, -1, 1 );
		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();

		glBegin(GL_QUADS);
			glTexCoord2f(0.0f, 0.0f); glVertex2i( 0, 0 );
			glTexCoord2f(1.0f, 0.0f); glVertex2i( g_dwWidth, 0 );
			glTexCoord2f(1.0f, 1.0f); glVertex2i( g_dwWidth, g_dwHeight );
			glTexCoord2f(0.0f, 1.0f); glVertex2i( 0, g_dwHeight );
		glEnd();

		SwapBuffers( g_hDC );

		// these openGL calls may be async 		
		// but the fake primary can't be left locked because the game might try to access it...
		// but by unlocking we risk the game drawing to the surface before glTexSubImage2D finishes. ( unlikely )
		// the surface data pointer won't become "invalid" because it is client allocated.
		This->dds1->lpVtbl->Unlock( This->dds1, NULL );
	}
}


